#lang racket

;;;; 6. cvičenie z predmetu 1-AIN-616 Symbolické programovanie a LISP
;;;;
;;;; Meno a priezvisko:
;;;; Verzia DrRacket: 6.4
;;;; Dátum: 22.3.2016
;;;; WWW: http://www.ii.fmph.uniba.sk/cl/view.php/courses/1-AIN-616-lisp/1516ls/?lang=sk 
;;;; Kontakt: Jan.Komara(at)fmph.uniba.sk 
;;;;
;;;; Úvodná poznámka. Toto cvičenie je venované týmto témam:
;;;; - univerzálna funkcia pre čiastočne rekurzívne funkcie,
;;;; - efektívne operácie na primitívne rekurzívnych indexoch.
;;;; Zadanie sa skladá z týchto súborov: 
;;;; - Súbory maux*.rkt obsahujú definície pomocných funkcií a predikátov. 
;;;; - Súbory ex*.rkt obsahujú úlohy, ktoré máte riešiť na tomto cvičení.
;;;;
;;;; Hodnotenie.
;;;; - Počet bodov za úlohy v tomto súbore: 6
;;;; - Počet bodov za prémiove úlohy v tomto súbore: 1
;;;; - Počet bodov za projektové úlohy v tomto súbore: 0
;;;; Odovzdať do 3.4.2016 (nedeľa).

;; [Lisp]

(require racket/undefined)
(define ... undefined)
(define (query f m n) (for/list ([i (in-range m (+ n 1))]) (list i (f i))))

;; Zopakovanie.

(define (iter f n x)
  (if (= n 0) x (iter f (- n 1) (f x))))

(define (each? p? xs)
  (if (null? xs) #t (and (p? (car xs)) (each? p? (cdr xs)))))

(define (natural-number? x)
  (exact-nonnegative-integer? x))

;; Literatúra.
;; [1] Hal Abelson and Jerry Sussman and Julie Sussman.
;;     Structure and Interpretation of Computer Programs.
;;     MIT Press, second edition, 1996. 

;; (define nil null)

;;; Kapitola. Čiastočne rekurzívne funkcie.

;;; Sekcia. Univerzálna funkcia pre čiastočne rekurzívne funkcie.

;;; Podsekcia. Aritmetizácia rekurzívnych termov a funkčných symbolov.

;; Konštruktory.

(define (Vt i) (list 'Vt i))               ; premenné
(define Zt (list 'Zt))                     ; konštanta 0
(define (Dt t1 t2 t3) (list 'Dt t1 t2 t3)) ; podmienkový výraz
(define (Ct t ts) (list 'Ct t ts))         ; kontrakcia argumentov
(define (At e ts) (list 'At e ts))         ; aplikácia funkcie

(define Sx (list 'Sx))           ; funkcia nasledovníka
(define Px (list 'Px))           ; funkcia predchodcu
(define (Fx n) (list 'Fx n))     ; funkčná premenná
(define (Lx n t) (list 'Lx n t)) ; definovaná funkcia

;; Predikáty.

(define (Vt? t) (equal? (list-ref t 0) 'Vt))
(define (Zt? t) (equal? (list-ref t 0) 'Zt))
(define (Dt? t) (equal? (list-ref t 0) 'Dt))
(define (Ct? t) (equal? (list-ref t 0) 'Ct))
(define (At? t) (equal? (list-ref t 0) 'At))

(define (Sx? e) (equal? (list-ref e 0) 'Sx))
(define (Px? e) (equal? (list-ref e 0) 'Px))
(define (Fx? e) (equal? (list-ref e 0) 'Fx))
(define (Lx? e) (equal? (list-ref e 0) 'Lx))

;; Selektory.

(define (Vt-index t) (list-ref t 1))
(define (Dt-arg1 t) (list-ref t 1))
(define (Dt-arg2 t) (list-ref t 2))
(define (Dt-arg3 t) (list-ref t 3))
(define (Ct-arg t) (list-ref t 1))
(define (Ct-args t) (list-ref t 2))
(define (At-fun t) (list-ref t 1))
(define (At-arg t) (list-ref t 2))

(define (Fx-arity e) (list-ref e 1))
(define (Lx-arity e) (list-ref e 1))
(define (Lx-body e) (list-ref e 2))

;; Údajový typ pre rekurzívne termy a funkčné symboly.

(define (TERM? t)
  (and (list? t)
       (>= (length t) 1)
       (or (and (Vt? t)
                (= (length t) 2)
                (natural-number? (Vt-index t)))
           (and (Zt? t)
                (= (length t) 1))
           (and (Dt? t)
                (= (length t) 4)
                (TERM? (Dt-arg1 t))
                (TERM? (Dt-arg2 t))
                (TERM? (Dt-arg3 t)))
           (and (Ct? t)
                (= (length t) 3)
                (TERM? (Ct-arg t))
                (TERM? (Ct-args t)))
           (and (At? t)
                (= (length t) 3)
                (FUNCTION? (At-fun t))
                (TERM? (At-arg t))))))

(define (FUNCTION? e)
  (and (list? e)
       (>= (length e) 1)
       (or (and (Sx? e)
                (= (length e) 1))
           (and (Px? e)
                (= (length e) 1))
           (and (Fx? e)
                (= (length e) 2)
                (natural-number? (Fx-arity e)))
           (and (Lx? e)
                (= (length e) 3)
                (natural-number? (Lx-arity e))
                (TERM? (Lx-body e))))))

;; Predikát recursive-function-symbol?(n,e) platí, ak číslo e je kódom n-árneho
;; rekurzívneho funkčného symbolu.

(define (recursive-function-symbol? n e)
  (define (is-recursive-function-symbol? n e)
    (or (and (Sx? e) (= n 1))
        (and (Px? e) (= n 1))
        (and (Lx? e)
             (= (Lx-arity e) n)
             (recursive-term? n (Lx-body e)))))
  (and (FUNCTION? e) (>= n 1) (is-recursive-function-symbol? n e)))

(define (recursive-term? n t)
  (define (is-recursive-term? n t)
    (or (and (Vt? t)
             (<= 1 (Vt-index t))
             (<= (Vt-index t) n))
        (Zt? t)
        (and (Dt? t)
             (is-recursive-term? n (Dt-arg1 t))
             (is-recursive-term? n (Dt-arg2 t))
             (is-recursive-term? n (Dt-arg3 t)))
        (and (At? t)
             (let* ([e (At-fun t)] [m (arity e)])
               (if (Fx? e)
                   (and (= m n)
                        (are-recursive-terms? n (At-arg t) n))
                   (and (>= m 1)
                        (recursive-function-symbol? m e)
                        (are-recursive-terms? n (At-arg t) m)))))))
  (define (are-recursive-terms? n ts k)
    (or (and (= k 1)
             (is-recursive-term? n ts))
        (and (> k 1)
             (Ct? ts)
             (is-recursive-term? n (Ct-arg ts))
             (are-recursive-terms? n (Ct-args ts) (- k 1)))))
  (and (TERM? t) (is-recursive-term? n t)))

;; Arita funkčných symbolov.

(define (arity e)
  (cond [(Sx? e) 1]
        [(Px? e) 1]
        [(Fx? e) (Fx-arity e)]
        [(Lx? e) (Lx-arity e)]))

;;; Podsekcia. Aritmetizácia redukcií.

;; Úloha. Zostrojte program, ktorý pre dané číslo vytvorí odpovedajúci numerál.

(define (number->numeral x) ...)

;; Úloha. Zostrojte program, ktorý pre danú n-ticu čísel vytvorí odpovedajúcu
;; n-ticu numerálov (skontrahovanú konštruktorom Ct).

(define (numbers->numerals n xs) ...)

;; Úloha. Zostrojte program, ktorý dekóduje numerál.

(define (numeral->number t) ...)

;; Úloha. Zostrojte program, ktorý overí, či term je numerál.

(define (numeral? t) ...)

;; Úloha. Zostrojte program, ktorý overí, či n-tica termov pozostáva z numerálov.

(define (numerals? n ts) ...)

;; Úloha. Implementujte operáciu substitúcie pre rekurzívne termy.

(define (subst t e rs) ...)

;; (4 body)
;; Úloha. Zostrojte program pre funkciu, ktorý realizuje jeden krok výpočtu.

(define (reduce t) ...)

;; Operácia eval(t) vyhodnotí daný uzavretý rekurzívny term.

(define (eval t)
  (if (numeral? t) t (eval (reduce t))))

;; Operácia apply(e,xs) aplikuje program e na argumenty xs.

(define (apply e xs)
  (eval (At e (numbers->numerals (arity e) xs))))

;; [Lisp] Táto časť slúži len pre ladenie programov.

(define (bounded-apply k e xs)
  (iter reduce k (At e (numbers->numerals (arity e) xs))))

;; Podsekcia. Rekurzívne indexy.

;; Definícia. Ak platí predikát recursive-function-symbol?(n,e), tak vravíme, že
;; číslo e je dobre vytvorený rekurzívny index n-árnej čiastočne rekurzívnej
;; funkcie.

;; (1 bod)
;; Úloha. Zostrojte dobre vytvorený rekurzívny index pre sčítanie.

(define ⌜+⌝ ...)

;; (FUNCTION? ⌜+⌝)
;; (recursive-function-symbol? 2 ⌜+⌝)
;; (bounded-apply 100 ⌜+⌝ (cons 3 2))
;; (numeral->number (apply ⌜+⌝ (cons 3 2)))

;; Úloha. Zostrojte dobre vytvorený rekurzívny index pre modifikované odčítanie.

(define ⌜-⌝ ...)

;; (FUNCTION? ⌜×⌝)
;; (recursive-function-symbol? 2 ⌜-⌝)
;; (bounded-apply 100 ⌜-⌝ (cons 3 2))
;; (numeral->number (apply ⌜-⌝ (cons 3 2)))

;; Úloha. Zostrojte dobre vytvorený rekurzívny index pre násobenie.

(define ⌜×⌝ ...)

;; (FUNCTION? ⌜×⌝)
;; (recursive-function-symbol? 2 ⌜×⌝)
;; (bounded-apply 100 ⌜×⌝ (cons 3 2))
;; (numeral->number (apply ⌜×⌝ (cons 3 2)))

;; (1 bod)
;; Úloha. Zostrojte dobre vytvorený rekurzívny index pre operáciu umocňovania.

(define ⌜exp⌝ ...)
  
;; (FUNCTION? ⌜exp⌝)
;; (recursive-function-symbol? 2 ⌜exp⌝)
;; (bounded-apply 100 ⌜exp⌝ (cons 3 2))
;; (numeral->number (apply ⌜exp⌝ (cons 3 2)))

;;; Podsekcia. Efektívne operácie na rekurzívnych indexoch.

;; Úloha. Zostrojte program pre funkciu xparam(e,x) takú, že platí:
;;   recursive-function-symbol?(2,e) → recursive-function-symbol?(1,xparam,(e,x))
;;   apply(xparam(e,x),y) = apply(e,cons(x,y))

(define (xparam e x) ...)

;; (FUNCTION? (xparam ⌜+⌝ 5))
;; (recursive-function-symbol? 1 (xparam ⌜+⌝ 5))
;; (query (λ (x) (numeral->number (apply (xparam ⌜+⌝ 5) x))) 0 10)

;; (prémia 1 bod)
;; Úloha. Zostrojte program pre funkciu xsmn-2-3(e,x1,x2) takú, že platí:
;;   recursive-function-symbol?(5,e) → recursive-function-symbol?(3,xsmn-3-5(e,x1,x2))
;;   apply(xsmn-2-3(e,x1,x2),cons(y1,cons(y2,y3)) = apply(e,cons(x1,cons(x2,cons(y1,cons(y2,y3)))))

(define (xsmn-2-3 e x1 x2) ...)

;; (query (λ (n) (FUNCTION? (xsmn-2-3 (Lx 5 (Vt 1)) n n))) 0 10)
;; (query (λ (n) (recursive-function-symbol? 3 (xsmn-2-3 (Lx 5 (Vt 1)) n n))) 0 10)
